.\" -*- nroff -*- 
.ds T1 Shell programming
.vl $Header: /home/jove/facad/trent/doc/groff/mo/RCS/shell.mo,v 1.1 91/08/27 19:11:06 trent Exp Locker: trent $
.ds T2 \*(Vl
.ds T3 CS Systems Group
.ds T4 EE Systems Group
.\" -----------------------------------------------------------------
.(c
.sz +4
.b "Shell programming"
.sz -2
.sp 2
.i "CS/EE Systems Group"
Workshop Lecture
.sp 2
Trent A. Fisher
.sz -2
.)c
.bp
.\" -----------------------------------------------------------------
.hd "Why Shell programming?"
.ls 2
.uh Advantages
.bu
Portability (no need to recompile)
.bu
Speed of writing/testing
.bu
Ease of pipelining information
.bu
No data types
.sp 3
.lp
.b Disadvantages
.bu
No complex data types.  (Arrays can be faked.)
.bu
Some operations difficult or impossible.
.bu
Can be much slower than C program.
.bu
No data types
.br
.ls 1
.bp
.\" -----------------------------------------------------------------
.hd "Why Bourne Shell?"
.bu
Lowest common denominator
.(q
When aiming for the common denominator, be prepared for the occasional
division by zero.
                --Jim des Rivieres
.)q
.sp
.bu
Size
.(l
text    data    bss     dec   
28672   4096    0       32768  /bin/sh
106496  8192    7632    122320 /bin/csh
126976  8192    4188    139356 /bin/ksh
188416  53248   5168    246832 /usr/local/bin/bash
.)l
.sp
.bu
Portability
.sp
.bu
More flexibility for programming
.sp
.bu
Less bugs
.bp
.\" -----------------------------------------------------------------
.hd "Purpose of A Shell"
.bu
Execute commands
.br
This means the shell has to create a new process
and have it run the desired binary.
.sp 2
.in +2i
.sz -2
.PS
arrow down
Fork: box "fork()"
arrow down 3i from last box.s
Wait: box "wait()"
arrow down .75i right 1i from Fork.e
box "exec()" with .n at last arrow.end
arrow down 1i from last box.s
box "new" "program"
arrow from last box.s to Wait.r
.PE
.in -2i
.sz +2
.sp
.bu
If a command is run in the background (i.e.
.c twm& )
the wait call is not done.
.bp
.\" -----------------------------------------------------------------
.hd "What the Shell does with a Command"
.bu
Break command line into arguments
For example, the command
.(l
gcc -g -Wall slartibartfast.c
.)l
would be broken up like this:
.sp
.sz -2
.in +1i
.c
.PS
box invis wid .5i "argv"
arrow right
A: box ht .3i wid .3i with .w at last arrow.end
{ arrow right from last box.c; box wid .5i ht .25i "gcc" }
B: box ht .3i wid .3i with .n at A.s
{ arrow right from last box.c; box wid .4i ht .25i "-g" }
C: box ht .3i wid .3i with .n at B.s
{ arrow right from last box.c; box wid .7i ht .25i "-Wall" }
D: box ht .3i wid .3i with .n at C.s
{ arrow right from last box.c; box wid 2i ht .25i "slartibartfast.c" }
E: box ht .3i wid .3i with .n at D.s "0"
.PE
.sz +2
.r
.sp
.bu
Find binary on the system
.sp
.bu
Execute the command
.sp
.bu
Supply the return status
.br
A status of 0 means the program worked normally,
anything else means something went wrong.
.bp
.\" -----------------------------------------------------------------
.hd "Other Shell Services"
.bu
Redirection
.sp
.bu
Piping
.sp
.bu
Filename globbing (meta-characters)
.sp
.bu
Builtin operations
.bp
.\" -----------------------------------------------------------------
.hd "Redirection"
.bu
I/O streams
.br
Every process has three default I/O streams:
stdin, stdout, stderr.
These streams are supplied by the parent process.
This means that a shell can change where they actually go
for any program they run
without the program's knowledge.
.br
Forms:
.sp
.po +.25i
.bu
Redirection
.br
.c "cat < x > y"
.sp
.bu
Append
.br
.c "cat x >> y"
.sp
.bu
By descriptor
.br
.c "gcc -Wall x.c 2> errs"
.sp
.bu
Merging streams
.br
.c "gcc -Wall x.c > output 2>&1"
.sp
.bu
Closing streams
.br
\fC<&-\fP	close stdin
.br
\fC>&-\fP	close stdout
.br
.po -.25i
.bp
.\" -----------------------------------------------------------------
.hd "Redirection, continued"
.bu
Here documents
.br
Read from script (here) until a keyword is encountered
.sp
.(l
cat << SPAM > x
How to identify different trees
      from quite a long way away.
Number one...
The Larch.
The...
Larch.
SPAM
.)l
An excellent example of the use of here documents is
.c shar .
.bp
.\" -----------------------------------------------------------------
.hd "Piping"
.bu
Connect
.i stdout
of command on the left
to the
.i stdin
of the command on the right.
.i Stderr
is unaffected.
.sp
The command line
.(l
cat /etc/passwd | sort -t: +2 | less
.)l
builds a pipeline like this:
.br
.sz -4
.PS
arrow "stdin" above
box ht .3i wid 1.1i "cat /etc/passwd"
{ line down .2i from last box.s; box invis ht .15i "stderr"; arrow down .3i}
arrow "stdout" above
box invis wid .1i "|"
arrow "stdin" above
box ht .3i wid .9i  "sort -t: +2"
{ line down .2i from last box.s; box invis ht .15i "stderr"; arrow down .3i}
arrow "stdout" above
box invis wid .1i "|"
arrow "stdin" above
box ht .3i wid .6i "less"
{ line down .2i from last box.s; box invis ht .15i "stderr"; arrow down .3i}
arrow "stdout" above
.PE
.sz +4
.bp
.\" -----------------------------------------------------------------
.hd "Filename Globbing"
.ip *
Matches any file (except ones starting with `.').
.sp
.ip ?
Matches any single character (except `.').
.sp
.ip [...]
Matches any one of the characters enclosed.
A pair of characters separated by `-'
matches any character lexically between the pair.
.sp 2
Examples:
.(l
$ ls -A
\&.bashrc .cshrc  README  x   xxx    y    z
$ echo *
README x xxx y z
$ echo ?
x y z
$ echo .*
\&. .. .bashrc .cshrc
$ echo [xyz]
x y z
$ echo [xyz]*
x xxx y z
$ echo .[a-z]*
\&.bashrc .cshrc
.)l
.bp
.\" -----------------------------------------------------------------
.hd "Builtin operations"
.bu
Flow control
.br
If-then-else, loops, and case statements
.sp 2
.bu
Convenience builtins
.br
Some utilities are built into the shell
for efficiency.
Commonly these are
.i test ,
.i echo ,
.i time ,
and others.
.sp 2
.bu
One utility
.bi must
be built into the shell:
.c cd .
.sp 2
.bu
Another utility which is built into the shell is
.c exec ,
it runs a process by overlaying it on top of the current one
(effectively destroying the current process).
.bp
.\" -----------------------------------------------------------------
.hd "Shell scripts"
.bu
The shell can read commands in an interactive fashion from a terminal,
or it can read a sequence of commands from a file.
There are several ways to do this (assuming that the file
.c njorl
is a file contains some shell commands):
.(l
sh njorl
.)l
.bu
The other way is to make the file executable:
.(l
chmod +x njorl
.)l
.lp
The first line of the script indicates what type of script it is.
The best way to specify what shell to use is the
.c #!
notation.
The first line of an executable file can contain a
.c #!
followed by the
.i absolute
pathname of the shell, i.e.
.c #!/bin/sh .
.lp
On some systems the first character may also specify what type of
script it is.
.sp
.bu
A colon indicates a bourne shell script.
.sp
.bu
A pound sign (
.c #)
without a trailing
.c !
indicates a csh script.
.sp
.bu
Yet other systems (mainly old System V machines)
recognize none of these and blindly run your script
under the same type of shell as your login shell.
.bp
.\" -----------------------------------------------------------------
.hd "Variables"
.lp
Identifiers for shell variables are similar to other languages:
an alphabetic character followed by alphabetic, numeric
or underscore characters.
.lp
Shell variables are all strings,
although they can be used as numbers.
.sp
.uh "Builtin Variables"
.lp
The shell maintains a number of variables which are built into the
shell, and the values are determined by the shell.
.ip ?
The value returned by the last executed command in decimal.
.ip $
The process number of this shell.
.ip !
The process number of the last background command invoked.
.ip -
Options supplied to the shell on invocation or by
.c set
.ip 0
Name of the current shell script, or the name of the shell.
.ip #
The number of command line arguments in decimal.
.ip [1-9]
The command line arguments to the current shell script.
.ip *
All the command line arguments.
.ip @
Similar to, but subtly different from `*'.
.bp
.\" -----------------------------------------------------------------
.hd "Variables, Part II"
.lp
Other shell variables have special meaning to the shell.
These variables are used for mainly configuration,
a few are set when the shell is invoked.
.sp 2
.ip PATH
The search path for commands.
It usually contains at least:
.(l
:/bin:/usr/bin:/usr/ucb:/usr/local/bin:/usr/games:
.)l
.ip PS1
.ip PS2
Prompt strings.  They default to `$ ' and `> ', respectively.
.sp
.ip USER
Your login name
.sp
.ip HOME
Your home directory.
The default argument for
.c cd .
.sp
.ip MAIL
Name of your mail file.
Usually,
.c /usr/spool/mail/$USER
.sp
.ip IFS
Internal field separators, normally space, tab, and newline.
.bp
.\" -----------------------------------------------------------------
.hd "Bride of Variables"
.uh Assignment
.br
Assigning a value to a variable is much like other programming
languages, using an `=' to indicate assignment.
For example:
.(l
tree=larch
meal="spam eggs sausage and spam"
.)l
.sp
.uh Gotchas
.br
Due to the nature of the shell you have to be careful about spaces.
For example:
.(l
$ tree = larch
tree: not found
$ tree= larch
larch: not found
$ meal=spam eggs sausage spam
eggs: not found
.)l
.bp
.\" -----------------------------------------------------------------
.hd "Son of Variables"
.uh Interpolation
.sp
.bu
To interpret a variable precede the identifier with 
.c `$' .
For example:
.(l
$ echo $PATH
/usr/local/gnu/bin:/home/rigel/rye/trent/bin:
/usr/ucb:/bin:/usr/bin:/usr/local:
/usr/local/bin:/usr/games:/usr/etc:
/usr/local/etc:/etc:/usr/bin/X11:
$ echo $$
21167
$ echo $-
s
.)l
.sp
.bu
You may also surround the identifier with braces (
.c `{}' ).
This is useful if you want to follow the variable immediately
with other characters.  For Example:
.(l
$ a1=quotes
$ a2=monty
$ echo ${a1}_$a2
quotes_monty
$ echo $a1_$a2
monty
.)l
.bp
.\" -----------------------------------------------------------------
.hd "The Variables that Ate Tokyo"
.uh "Parameter substitution"
.lp
Determine if a shell variable is set,
and perform various actions accordingly.
.sp
.bu
${identifier:-word}
.br
If the variable is set and is nonnull,
substitute its value;
otherwise substitute `word'.
.sp
.bu
${identifier:=word}
.br
If the variable is not set or is null set it to word;
the value of the parameter is substituted.
.sp
.bu
${identifier:?word}
.br
If parameter is set and is nonnull,
substitute its value;
otherwise, print word and exit from the shell.
If word is omitted, the message
.c `parameter null or not set'
is printed.
.sp
.bu
${identifier:+word}
.br
If parameter is set and is nonnull,
substitute word;
otherwise substitute nothing.
.lp
If the colon is omitted, the shell will only test if the variable
is unset as opposed to unset or null.
.sp
.uh Gotchas
.lp
Many shells have more of these,
and some don't have all of them.
.bp
.\" -----------------------------------------------------------------
.hd "Night of the Living Variables"
.uh "Environment Variables"
.lp
Normally shell variables are not passed to children.
However, they can be put into the
.i environment
via the
.c export
command,
which will give children processes access to them.
.sp
.(l
$ a=var-a
$ b=var-b
$ export b
$ echo $a $b
var-a var-b
$ sh
$ echo $a $b
var-b
.)l
.sp
Many builtin shell variables must be
.c export ed,
such as
.c PATH ,
.c PS1 ,
.c PS2 ,
and
.c HOME .
.bp
.\" -----------------------------------------------------------------
.hd "Control structures"
.lp
The shell provides a number of control structures.
However, most of them have deviations which do not exist in other
languages.
.lp
The structures are:
.bu
For loops
.bu
Case statements
.bu
While loops
.bu
If then else statements
.bu
Logical or
.bu
Logical and
.bp
.\" -----------------------------------------------------------------
.hd "For loop"
.lp
Unlike for loops in other languages
(possibly like pascal or ada).
.sp 2
.(l
for i in *
do
    compress $i &
done
.)l
.sp
.lp
Anything may be given after the
.c in ,
for example:
.(l
for i in spam eggs bacon sausage
do
	echo $i
done
.)l
.sp
.lp
The command line (\c
.c $* )
is implicit with the for loop.
.sp
.(l
for i
do
    echo  $i
done
.)l
.bp
.\" -----------------------------------------------------------------
.hd "Case Statement"
.lp
The case statement is similar to case statements in other languages
except it allows globbing and logical expressions.
.(l
for i
do
	case $i in
        	\(mi[ocs]) ... ;;
		\(miverbose|\(midebug)
			debug=1 ;;
        	\(mi*)	echo unknown flag $i ;;
		*.c)	/lib/c0 $i ... ;;
		\e?)	echo what\e? ;;
		*)	echo unexpected argument $i ;;
	esac
done
.)l
.bp
.\" -----------------------------------------------------------------
.hd "A Digression on Test"
.lp
An understanding of the command
.c test
is vital for what follows.
.lp
The
.c test
command takes its arguments as logical expressions
and returns true or false via the programs exit status.
.lp
.c Test
can also be called via the
.c [
command, the only difference is that in this form it must
have a trailing
.c ] .
.uh "File Tests"
.sp
.ip "-r file"
true if the file exists and is readable.
.sp
.ip "-w file"
true if the file exists and is writable.
.sp
.ip "-f file"
true if the file exists and is not a directory.
.sp
.ip "-d file"
true if the file exists and is a directory.
.sp
.ip "-s file"
true if the file exists and has a size greater than zero.
.sp
.ip "-t [ fildes ]"
true if the open file whose file descriptor number
is fildes (1 by default) is associated with a terminal device.
.bp
.\" -----------------------------------------------------------------
.hd "A Digression on Test, continued"
.uh "String Tests"
.sp
.ip "-z s1"
true if the length of string s1 is zero.
.sp
.ip "-n s1"
true if the length of the string s1 is nonzero.
.sp
.ip "s1 = s2"
true if the strings s1 and s2 are equal.
.sp
.ip "s1 != s2"
true if the strings s1 and s2 are not equal.
.sp
.ip "s1"
true if s1 is not the null string.  This will only
work if no other primitives are used in expr.
.uh "Numeric operations"
.sp
.ip "n1 -eq n2"
true if the integers n1 and n2 are algebraically
equal.  Any of the comparisons -ne, -gt, -ge, -lt,
or -le may be used in place of -eq.
.bp
.\" -----------------------------------------------------------------
.hd "A Digression on Test, continued"
.uh "Logical Operators"
.sp
.ip !
unary negation operator
.sp
.ip -a
binary and operator
.sp
.ip -o
binary or operator
.sp
.ip "( expr )"
parentheses for grouping.
.lp
.bu 
-a has higher precedence than -o.
.uh Gotchas
.sp
.bu
Notice that all the operators and flags are separate arguments to
test.
.sp
.bu
Notice also
that parentheses are meaningful to the Shell and must be
escaped.
.bp
.\" -----------------------------------------------------------------
.hd "If statement"
.lp
The
.c if
statement is similar to other programming languages,
except instead of a logical expression,
it takes a command and evaluates its exit status.
.(l
if grep -s root /etc/passwd
then
	echo ok
else
	echo we got trouble
fi
.)l
.lp
Or using the
.c test
command:
.(l
if [ -w / ]
then
	echo bad news
fi
.)l
You can also chain
.c  else-if 's
via the
.c elif
construct
.(l
if [ $x -gt 0 ]
then
	echo positive
elif [ $x -lt 0 ]
then
	echo negative
else
	echo zero
fi
.)l
.uh Gotcha
.br
Note that the shell concept of true or false is reverse of
C and other languages:  0 is true, anything else is false.
.bp
.\" -----------------------------------------------------------------
.hd "While loop"
.lp
Much like the
.c if ,
The while loop takes a command as its condition
and evaluates the exit status to determine whether to continue to
loop.
.(l
x=10
while [ $x -ge 0 ]
do
	echo -n "$x, "
	x=`expr $x - 1`		# decrement $x
done
echo boom
.)l
.bp
.\" -----------------------------------------------------------------
.hd "Logical Operators"
.lp
The shell will allow you to chain commands together
logical operators.
For example:
.(l
if [ -w .rhosts ]
then
	chmod go-w .rhosts
fi
.)l
.lp
Can be rewritten as
.(l
test -w .rhosts && chmod go-w .rhosts
.)l
.lp
The second command is executed only if the first command succeeds.
The
.c or
operator works similarly
.(l
test ! -w .rhosts || chmod go-w .rhosts
.)l
.lp
The second command is executed only if the first one
.i fails .
.bp
.\" -----------------------------------------------------------------
.hd "Command grouping"
.lp
There are two ways to group commands:
.(l
{ commands ; }
.)l
or
.(l
( commands )
.)l
.lp
The former merely executes the commands in sequence,
the latter executes the commands in a subshell.
A simple example is:
.(l
(cd monty; rm chapman)
.)l
In this case the
.c cd
will not affect the current shell.
.lp
A more useful example is:
.(l
(cd /src/gnu/groff; tar cvf - ) | tar xvf -
.)l
.sp
.bu
In both cases the commands should be separated by semi-colons.
.bp
.\" -----------------------------------------------------------------
.hd "Quoting"
.lp
Quotes are taken care of before the command is broken
into arguments.
.in +1i
.TS
box,tab(:);
l | c c c c c c
_ = = = = = = =
l || c c c c c c.
:\e:$:*:\`:":\'
\':n:n:n:n:n:t
\`:y:n:n:t:n:n
": y:y:n:y:t:n
.TE
.in +.5i
t	terminator
.br
y	interpreted
.br
n	not interpreted
.in -1.5i
.sp 2
.uh "Command Substitution"
.sp
.lp
The back-quotes (or back-ticks) indicate that the contents of
the quotes should be executed as a command
and the output from the command is substituted.
.(l
$ echo `date`
Tue Aug 27 14:06:07 PDT 1991
.)l
.sp
A more useful (and more dangerous) example is:
.(l
files=`find . -name '*.Z' -print`
for i in $files
do
	uncompress $i&
done
.)l
.bp
.\" -----------------------------------------------------------------
.hd "Fault handling"
.lp
The shell has the ability to have signal handlers.
This way, a script can respond to an interrupt
and perform cleanup.
.lp
The
.c trap
command allows certain operations to be performed
in response to certain signals.
.(l
trap 'rm /tmp/$$; exit' 2
.)l
.lp
If the
.c exit
is not done the shell will resume execution where the signal was
received.
.lp
Signals can also be ignored using
.c trap :
.(l
trap '' 1 2 3 15
.)l
.bp
.\" -----------------------------------------------------------------
.hd "Other Utilities"
.lp
Shell programming depends on many peripheral programs
to do various tasks.
.ip echo
.sp
.ip test
.sp
.ip basename
Removes a specified suffix from a file, for example:
.(l
cc \$1
mv a.out `basename \$1 .c`
.)l
.sp
.ip expr
evaluate mathematical expressions.
Also, some string operations
.sp
.ip pwd
Print the current working directory.
.sp
.ip whoami
Prints the login name of the current effective UID.
.sp
.ip "who am i"
Prints the login name of the person logged in on the current tty.
.sp
.ip tee
Pipe fitting.
.sp
.ip yes
Print a given word (the letter `y' by default) infinitely.
This is useful for commands that ask stupid questions.
.sp
.ip true
.ip false
Return status of 1 or 0.
This is useful for infinite while loops.
.sp
.ip awk
Pattern scanning, manipulating tabular data
.sp
.ip sed
Editing
.bp
.\" -----------------------------------------------------------------
.hd "AWK"
.lp
Awk is a simple programming language which is suited to manipulation
of tabular data.
.bp
.\" -----------------------------------------------------------------
.hd Sed
.lp
Stream editor
.bp
.\" -----------------------------------------------------------------
.hd "Comparisons of Shells"
.bu
Bourne shell is more generous about newlines in () and quotes.
The C-shell in particular does not allow newlines,
which make programming more difficult
.sp
.bu
Tilde expansion.
.br
Most modern shells use the tilde character for several things:
.c ~/dir
expands the
.c `~'
to your home directory,
.c ~user
expands to
.c user 's
home directory.
.sp
.bu
More parameter substitutions
.br
In fact the POSIX shell standard specifies many more,
which effectively eliminates the need for
.c basename
.sp
.bu
Aliases
.br
Some shells allow command aliases,
some allow argument interpolation (csh) others don't (bash)
.sp
.bu
Functions
.br
Many modern shells have shell functions,
which operate like functions in most other programming languages.
Usually, the ability to create local variables is also given.
